package com.paxos.absledge.testutil

import com.google.common.truth.Truth.assertThat
import com.google.common.truth.extensions.proto.ProtoTruth
import com.nhaarman.mockitokotlin2.any
import com.nhaarman.mockitokotlin2.argumentCaptor
import com.nhaarman.mockitokotlin2.clearInvocations
import com.nhaarman.mockitokotlin2.doAnswer
import com.nhaarman.mockitokotlin2.mock
import com.nhaarman.mockitokotlin2.times
import com.nhaarman.mockitokotlin2.verify
import com.paxos.absledge.external.LedgerTestWallet
import com.paxos.absledge.manager.AccountsManager
import com.paxos.absledge.manager.BalancesManager
import com.paxos.absledge.manager.DepositWithdrawalGenerator
import com.paxos.absledge.manager.LedgerManager
import com.paxos.absledge.manager.SettlementSender
import com.paxos.absledge.manager.SettlementUpdateProcessor
import com.paxos.absledge.manager.SettlementsManager
import com.paxos.absledge.model.EscrowType
import com.paxos.absledge.model.OfflineSettlement
import com.paxos.absledge.model.PendingActionSettlement
import com.paxos.absledge.model.PendingDepositSettlement
import com.paxos.absledge.model.ReservedSettlement
import com.paxos.absledge.model.SettledSettlement
import com.paxos.absledge.model.SettlementDTO
import com.paxos.absledge.model.SettlementStore
import com.paxos.absledge.persistence.MultilateralGroupStore
import com.paxos.absledge.persistence.NetSettlementStore
import com.paxos.db.DEFAULT_DECIMAL_SCALE
import com.paxos.eltnam.AssetKey
import com.paxos.eltnam.DateRange
import com.paxos.eltnam.absoluteLedgeCommodities
import com.paxos.eltnam.util.CurrentDateTime
import com.paxos.eltnam.walletCommodities
import com.paxos.jooq.absoluteledge.tables.references.MULTILATERAL_GROUPS
import com.paxos.jooq.absoluteledge.tables.references.MULTILATERAL_GROUP_ACTIONS
import com.paxos.jooq.absoluteledge.tables.references.NET_SETTLEMENTS
import com.paxos.jooq.absoluteledge.tables.references.SETTLEMENTS
import com.paxos.ledger.LedgerAccount
import com.paxos.ledger.testutil.accountGen
import com.paxos.multilateral.MultilateralGroupFilter
import com.paxos.multilateral.NetSettlement
import com.paxos.multilateral.PendingDepositNetSettlement
import com.paxos.multilateral.SettlementGroupId
import com.paxos.proto.absledge.LedgerBlock
import com.paxos.proto.settlements.Settlement
import com.paxos.proto.settlements.SettlementUpdate
import com.paxos.proto.settlements.SettlementUpsert.AffirmationChange
import com.paxos.testutils.assertTableSizeChange
import com.paxos.testutils.genDateTime
import com.paxos.testutils.testJooqTransaction
import io.kotest.assertions.fail
import org.joda.time.DateTime
import org.koin.core.Koin
import org.koin.core.module.Module
import java.math.BigDecimal
import java.time.LocalDate
import java.util.UUID

abstract class SettlementsSpec(
    body: SettlementsSpec.(Koin) -> Unit,
    modules: List<Module> = defaultTestDatabaseModules,
    properties: Map<String, String> = mapOf(),
    removeModules: List<Module> = emptyList(),
    addModules: List<Module> = emptyList(),
) : AbsLedgeDatabaseSpec({}, modules, properties, removeModules, addModules) {
    init {
        body(koin)
    }

    val currentDateTime: CurrentDateTime = koin.get()
    val accountsManager: AccountsManager = koin.get()
    val ledger: LedgerManager = koin.get()
    val balances: BalancesManager = koin.get()

    val settlementStore: SettlementStore = koin.get()
    val netSettlementStore: NetSettlementStore = koin.get()
    val multilateralGroupStore: MultilateralGroupStore = koin.get()
    val mockTransferRequestGenerator = mock<DepositWithdrawalGenerator> {
        on { maybeGenerateDepositRequest(any<PendingDepositSettlement>(), any()) }.doAnswer {
            it.arguments[0] as PendingDepositSettlement
        }
        on { maybeGenerateDepositRequest(any<PendingDepositNetSettlement>(), any()) }.doAnswer {
            it.arguments[0] as PendingDepositNetSettlement
        }
    }

    fun Settlement.newRootId(): Settlement =
        toBuilder().setRootId(UUID.randomUUID().toString()).build()

    private fun newSettlementSender(sendSettlementUpdateToParticipant: SettlementUpdateProcessor) =
        SettlementSender(koin.get(), koin.get(), koin.get(), koin.get(), koin.get(), sendSettlementUpdateToParticipant)

    fun newSettlementsManager(sendSettlementUpdateToParticipant: SettlementUpdateProcessor) =
        SettlementsManager(koin.get(),
            koin.get(),
            koin.get(),
            koin.get(),
            koin.get(),
            koin.get(),
            koin.get(),
            mockTransferRequestGenerator,
            koin.get(),
            newSettlementSender(sendSettlementUpdateToParticipant),
            koin.get())

    fun expectNoUpdatesSettlementsManager() =
        newSettlementsManager { _, _ ->
            fail("Expected no participant settlement updates to be sent")
        }

    fun createAccounts(
        n: Int,
        multilateralSettlementEnabled: Boolean = false,
        ledgerAssets: Set<AssetKey> = absoluteLedgeCommodities,
    ): List<LedgerAccount> =
        accountGen(multilateralSettlementEnabled = multilateralSettlementEnabled, ledgerAssets = ledgerAssets)
            .samples().take(n).map { sample ->
                val account = sample.value
                accountsManager.upsertAccount(account)
                assertThat(account.hasWalletProfile).isTrue()
                account
            }.toList()

    data class ParticipantSettlementUpdate(val update: SettlementUpdate, val participantId: String)

    fun affirmationChangeAndAssert(
        groupId: UUID,
        approved: List<Settlement>,
        waiting: List<Settlement>,
        expectedNewSettlements: Int = approved.size + waiting.size,
        expectedPendingAction: Int = 0,
        expectedPendingDeposit: Int = 0,
        expectedNewDeposits: Int = expectedPendingDeposit,
        expectedReserved: Int = 0,
        expectedOfflineApproved: Int = 0,
        expectedOfflineUnapproved: Int = 0,
        expectedUpdates: Int = 2,
        expectedDeletedSettlements: Int = 0,
        expectedNewMultilateralGroup: Boolean = false,
        expectedNewMultilateralGroupActions: Int = 0,
    ) {
        val count = approved.size + waiting.size
        val allSettlements = approved + waiting

        val upsert = AffirmationChange.newBuilder()
            .setGroupId(groupId.toString())
            .addAllApproved(approved)
            .addAllWaiting(waiting)
            .build()

        val participantUpdates = mutableListOf<ParticipantSettlementUpdate>()
        val expectedSettlementProtos = allSettlements.map {
            val quantity = it.quantity.toBigDecimal()
            it.toBuilder()
                .setQuantity(quantity.setScale(DEFAULT_DECIMAL_SCALE).toPlainString())
                .setGroupId(groupId.toString())
                .build()
        }.toTypedArray()

        val halfExpectedUpdates = expectedUpdates / 2
        val settlementsManager = newSettlementsManager { update, participantId ->
            if (participantUpdates.size < halfExpectedUpdates && expectedNewMultilateralGroup) {
                // If the group has changed, we expect the first half of the updates to be clearing the old group
                assertThat(update.hasGroupUpdate()).isTrue()
                assertThat(update.groupUpdate.groupId).isEqualTo(groupId.toString())
                assertThat(update.groupUpdate.settlementsCount).isEqualTo(0)
            } else {
                assertThat(update.hasGroupUpdate()).isTrue()
                // if multilateral group is expected, the group id will match that
                // we won't know the group id ahead of time, so it is asserted after the change is processed
                if (!expectedNewMultilateralGroup) {
                    assertThat(update.groupUpdate.groupId).isEqualTo(groupId.toString())
                }
                assertThat(update.groupUpdate.settlementsCount).isEqualTo(expectedSettlementProtos.size)
                val fieldsToIgnore = listOfNotNull(
                    Settlement.ROOT_ID_FIELD_NUMBER,
                    Settlement.PENDING_DEPOSIT_AT_FIELD_NUMBER,
                    Settlement.RESERVED_AT_FIELD_NUMBER,
                    Settlement.SETTLED_AT_FIELD_NUMBER,
                    Settlement.OFFLINE_AT_FIELD_NUMBER,
                    if (expectedNewMultilateralGroup) Settlement.GROUP_ID_FIELD_NUMBER else null,
                )
                ProtoTruth.assertThat(update.groupUpdate.settlementsList)
                    .ignoringFields(fieldsToIgnore)
                    .containsExactly(*expectedSettlementProtos)
            }
            participantUpdates.add(
                ParticipantSettlementUpdate(
                    update,
                    participantId
                )
            )
        }

        assertTransferRequestGeneratorNewDepositCalled(expectedNewDeposits) {
            db.assertTableSizeChange(expectedNewMultilateralGroupActions, MULTILATERAL_GROUP_ACTIONS) {
                val multilateralGroupChange = if (expectedNewMultilateralGroup) 1 else 0
                db.assertTableSizeChange(multilateralGroupChange, MULTILATERAL_GROUPS) {
                    db.assertTableSizeChange(expectedNewSettlements + expectedDeletedSettlements, SETTLEMENTS) {
                        settlementsManager.processAffirmationChange(upsert)
                    }
                }
            }
        }

        // if we're expecting the settlements to be moved to a multilateral group, fetch the new group id
        val expectedFinalGroupId = if (expectedNewMultilateralGroup) {
            db.testJooqTransaction {
                val settlementDate = LocalDate.parse(allSettlements[0].settlementDate)
                val filter = MultilateralGroupFilter(
                    settlementDateRange = DateRange(start = settlementDate, end = settlementDate),
                )
                val groups = multilateralGroupStore.getGroups(this, filter, currentDateTime.get())
                assertThat(groups).hasSize(1)
                // check there are no settlements with the old group id
                val settlementsWithOldGroupId = settlementStore.getByGroupId(this, groupId)
                assertThat(settlementsWithOldGroupId).isEmpty()
                // the new multilateral group id
                groups[0].attributes.id.value
            }
        } else groupId

        assertThat(participantUpdates).hasSize(expectedUpdates)
        if (participantUpdates.isNotEmpty()) {
            val (firstHalf, secondHalf) = participantUpdates
                .map { it.update.groupUpdate.groupId }
                .withIndex()
                .partition { (index, _) ->
                    index < halfExpectedUpdates
                }
            val firstHalfGroupIds = firstHalf.mapTo(mutableSetOf()) { (_, value) -> value }
            val secondHalfGroupIds = secondHalf.mapTo(mutableSetOf()) { (_, value) -> value }
            // the old settlements (first half) will be deleted
            assertThat(firstHalfGroupIds).containsExactly(groupId.toString())
            assertThat(secondHalfGroupIds).containsExactly(expectedFinalGroupId.toString())
        }

        db.testJooqTransaction {
            val settlements = settlementStore.getByGroupId(this, expectedFinalGroupId)
            assertThat(settlements).hasSize(count)
            assertThat(settlements.filterIsInstance<PendingActionSettlement>()).hasSize(expectedPendingAction)
            assertThat(settlements.filterIsInstance<PendingDepositSettlement>()).hasSize(expectedPendingDeposit)
            assertThat(settlements.filterIsInstance<ReservedSettlement>()).hasSize(expectedReserved)
            val offlineSettlements = settlements.filterIsInstance<OfflineSettlement>()
            assertThat(offlineSettlements.filter { it.approvedForSettlement }).hasSize(expectedOfflineApproved)
            assertThat(offlineSettlements.filter { !it.approvedForSettlement }).hasSize(expectedOfflineUnapproved)
            assertThat(settlementStore.getUnsent(this, groupId)).isEmpty()
        }
    }

    fun assertTransferRequestGeneratorNewDepositCalled(invocationCount: Int = 1, block: () -> Unit) {
        val args = argumentCaptor<PendingDepositSettlement>()
        clearInvocations(mockTransferRequestGenerator)
        block()
        verify(mockTransferRequestGenerator, times(invocationCount)).maybeGenerateDepositRequest(args.capture(), any())
    }

    fun assertTransferRequestGeneratorNewWithdrawalCalled(invocationCount: Int, block: () -> Unit) {
        val args = argumentCaptor<SettledSettlement>()
        clearInvocations(mockTransferRequestGenerator)
        block()
        verify(mockTransferRequestGenerator, times(invocationCount)).maybeGenerateWithdrawalRequest(args.capture(),
            any())
    }

    fun saveSettlement(settlement: SettlementDTO, blockId: UUID? = null, createdAt: DateTime = currentDateTime.get()) =
        db.testJooqTransaction {
            settlementStore.save(this, settlement, createdAt, blockId)
        }

    fun List<SettlementDTO>.save() {
        db.testJooqTransaction {
            val ids = mapNotNull {
                settlementStore.save(this, it, currentDateTime.get())
                settlementStore.getByRootId(this, it.attributes.rootId)?.attributes?.id
            }
            settlementStore.saveUpdateSentAt(this, ids, genDateTime())
        }
    }

    fun saveNetSettlement(
        settlement: NetSettlement,
        blockId: UUID? = null,
        createdAt: DateTime = currentDateTime.get(),
    ) =
        db.testJooqTransaction {
            netSettlementStore.save(this, settlement, createdAt, blockId)
        }

    private fun createMultilateralGroup(
        settlementDate: LocalDate,
        accounts: List<LedgerAccount>,
        expiryTime: DateTime,
    ) =
        createMultilateralGroup(settlementDate, accounts.mapTo(mutableSetOf()) { it.accountId }, expiryTime)

    fun createMultilateralGroup(settlementDate: LocalDate, accounts: Set<String>, expiryTime: DateTime) =
        db.testJooqTransaction {
            multilateralGroupStore.getOrCreate(this, settlementDate, accounts, currentDateTime.get()) {
                expiryTime
            }
        }

    fun createOpenMultilateralGroup(settlementDate: LocalDate, accounts: List<LedgerAccount>) =
        createMultilateralGroup(settlementDate, accounts, currentDateTime.get().plusHours(1))

    fun expectedSettlementProtos(group: List<SettlementDTO>) =
        group.map {
            it.toProtoBuilder().setQuantity(it.attributes.quantity.setScale(DEFAULT_DECIMAL_SCALE).toPlainString())
                .build()
        }.toTypedArray()

    fun getUnsentSettlements() =
        db.testJooqTransaction {
            settlementStore.getUnsent(this)
        }

    private fun mintLedgerBalance(account: LedgerAccount, asset: AssetKey, quantity: BigDecimal) {
        val mint = TestActions.fromAccount(account) {
            it.value.assetKey == asset.toProto()
        }.mint.derive(quantity = quantity.toPlainString())
        val block = LedgerBlock.newBuilder().addActions(mint).build()
        ledger.append(block)
    }

    fun mintSufficientBalance(wallet: LedgerTestWallet, account: LedgerAccount, settlement: SettlementDTO) {
        mintBalance(wallet, account, settlement.attributes.asset, settlement.attributes.quantity.abs())
    }

    fun mintBalance(wallet: LedgerTestWallet, payer: LedgerAccount, asset: AssetKey, quantity: BigDecimal) {
        if (asset in walletCommodities && !payer.useSwiftForUsd) {
            assertThat(payer.hasWalletProfile).isTrue()
            wallet.setBalance(payer.walletProfileId!!, asset, quantity)
        } else {
            mintLedgerBalance(payer, asset, quantity)
        }
        assertThat(balances.accountBalance(payer.accountId, asset)).isEqualToIgnoringScale(quantity)
    }

    fun testSettlementsJob(
        settlementDate: LocalDate,
        expectedRuns: Int,
        expectedNewSettlements: Int = 0,
        expectedPendingDepositCountRemaining: Int = 0,
        expectedNewNetSettlements: Int = 0,
        expectedPendingDepositNetCountRemaining: Int = 0,
        onUpdate: SettlementUpdateProcessor = { _, _ ->
            fail("Expected no settlement updates to be sent")
        },
    ) {
        val settlementsManager = newSettlementsManager(onUpdate)
        var runs = 0
        fun testSettlementsJob() {
            runs += 1
            if (runs > expectedRuns) {
                fail("Expected only $expectedRuns runs of the settlements job")
            }
            settlementsManager.runSettlementsJob {
                testSettlementsJob()
            }
        }

        db.assertTableSizeChange(expectedNewSettlements, SETTLEMENTS) {
            db.assertTableSizeChange(expectedNewNetSettlements, NET_SETTLEMENTS) {
                testSettlementsJob()
            }
        }

        assertThat(runs).isEqualTo(expectedRuns)

        assertThat(getUnsentSettlements()).isEmpty()
        db.testJooqTransaction {
            val pendingDepositCount = settlementStore.getPendingDeposit(this, settlementDate).size
            assertThat(pendingDepositCount).isEqualTo(expectedPendingDepositCountRemaining)

            val pendingDepositNetCount = netSettlementStore.getPendingDeposit(this, settlementDate).size
            assertThat(pendingDepositNetCount).isEqualTo(expectedPendingDepositNetCountRemaining)
        }
    }

    fun testSettlementsJobAndExpectReserved(
        settlementDate: LocalDate,
        expectedRuns: Int,
        expectedNewSettlements: Int,
        expectedPendingDepositCountRemaining: Int = 0,
        expectedNewNetSettlements: Int = 0,
        expectedPendingDepositNetCountRemaining: Int = 0,
        expectedReservedUpdates: Int,
        customAssertions: ((SettlementUpdate, String, Int, SettlementUpdateProcessor) -> Unit)? = null,
    ) {
        val participantUpdates = mutableListOf<ParticipantSettlementUpdate>()

        var updateCount = 0
        val defaultAssertions: SettlementUpdateProcessor = { update, _ ->
            assertThat(update.hasSingleUpdate()).isTrue()
            assertThat(update.singleUpdate.settlement.hasReservedAt()).isTrue()
        }
        testSettlementsJob(
            settlementDate,
            expectedRuns,
            expectedNewSettlements,
            expectedPendingDepositCountRemaining,
            expectedNewNetSettlements,
            expectedPendingDepositNetCountRemaining,
        ) { update, participantId ->
            customAssertions?.invoke(update, participantId, updateCount, defaultAssertions)
                ?: defaultAssertions(update, participantId)
            participantUpdates.add(ParticipantSettlementUpdate(update, participantId))
            updateCount += 1
        }

        val numberOfParticipants = participantUpdates.mapTo(mutableSetOf()) { it.participantId }.size
        assertThat(participantUpdates.size).isEqualTo(numberOfParticipants * expectedReservedUpdates)
    }

    fun testSettlementsJobAndExpectSettled(
        settlementDate: LocalDate,
        participants: Int = 2,
        expectedRuns: Int,
        expectedNewSettlements: Int,
        expectedNewNetSettlements: Int = 0,
        expectedPendingDepositCountRemaining: Int = 0,
        expectedPendingDepositNetCountRemaining: Int = 0,
        expectedReservedUpdates: Int,
        expectedSettledUpdates: Int,
        expectedSettledCount: Int,
        expectedOfflineCount: Int = 0,
        customAssertions: ((SettlementUpdate, String, Int, SettlementUpdateProcessor) -> Unit)? = null,
    ) {
        val participantUpdates = mutableListOf<ParticipantSettlementUpdate>()
        var updateCount = 0

        val defaultAssertions: SettlementUpdateProcessor = { update, _ ->
            if (updateCount < participants * expectedReservedUpdates) {
                assertThat(update.hasSingleUpdate()).isTrue()
                assertThat(update.singleUpdate.settlement.hasReservedAt()).isTrue()
            } else {
                assertThat(update.hasGroupUpdate()).isTrue()
                assertThat(update.groupUpdate.settlementsCount).isEqualTo(expectedSettledCount + expectedOfflineCount)
                val (settled, offline) = update.groupUpdate.settlementsList.partition {
                    it.hasReservedAt() && it.hasSettledAt()
                }
                assertThat(settled.size).isEqualTo(expectedSettledCount)
                assertThat(offline.size).isEqualTo(expectedOfflineCount)
                offline.forEach {
                    assertThat(it.hasOfflineAt()).isTrue()
                }
            }
        }

        testSettlementsJob(
            settlementDate,
            expectedRuns,
            expectedNewSettlements,
            expectedPendingDepositCountRemaining,
            expectedNewNetSettlements,
            expectedPendingDepositNetCountRemaining,
        ) { update, participantId ->
            customAssertions?.invoke(update, participantId, updateCount, defaultAssertions)
                ?: defaultAssertions(update, participantId)
            participantUpdates.add(ParticipantSettlementUpdate(update, participantId))
            updateCount += 1
        }

        val numberOfParticipants = participantUpdates.mapTo(mutableSetOf()) { it.participantId }.size
        assertThat(numberOfParticipants).isEqualTo(participants)
        assertThat(participantUpdates.size).isEqualTo(participants * expectedReservedUpdates + participants * expectedSettledUpdates)
    }

    fun assertNoBalanceOrEscrowChange(
        asset: AssetKey,
        groupId: SettlementGroupId,
        account: LedgerAccount,
        block: (() -> Unit),
    ) {
        val escrowType = EscrowType.fromGroupId(groupId)

        val balanceBefore = balances.accountBalance(account.accountId, asset)
        val escrowBefore = balances.escrowBalance(account.accountId, asset, escrowType)

        block()

        val balanceAfter = balances.accountBalance(account.accountId, asset)
        val escrowAfter = balances.escrowBalance(account.accountId, asset, escrowType)

        assertThat(balanceAfter).isEqualToIgnoringScale(balanceBefore)
        assertThat(escrowAfter).isEqualToIgnoringScale(escrowBefore)
    }
}
